''' 
Test the basic RotorPy environment. 
'''

from rotorpy.environments import Environment
from rotorpy.vehicles.multirotor import Multirotor
from rotorpy.vehicles.crazyflie_params import quad_params
from rotorpy.controllers.quadrotor_control import SE3Control
from rotorpy.trajectories.hover_traj import HoverTraj
from rotorpy.trajectories.circular_traj import ThreeDCircularTraj
from rotorpy.trajectories.lissajous_traj import TwoDLissajous
from rotorpy.trajectories.speed_traj import ConstantSpeed
from rotorpy.trajectories.minsnap import MinSnap
from rotorpy.wind.default_winds import NoWind, ConstantWind, SinusoidWind, LadderWind
from rotorpy.wind.dryden_winds import DrydenGust, DrydenGustLP
from rotorpy.wind.spatial_winds import WindTunnel
from rotorpy.sensors.imu import Imu
from rotorpy.sensors.external_mocap import MotionCapture
from rotorpy.world import World
from rotorpy.utils.postprocessing import unpack_sim_data

import numpy as np     
import matplotlib
matplotlib.use('Agg')  # use non-interactive backend for test            
import matplotlib.pyplot as plt     
from scipy.spatial.transform import Rotation 
import os

def test_basic_env():
    

    print("\nTesting basic environment setup...")

    # Create the world
    world = World.from_file(os.path.abspath(os.path.join(os.path.dirname(__file__),'..','rotorpy','worlds','double_pillar.json')))

    # An instance of the simulator can be generated as follows: 
    print("\tCreating the environment...")
    sim_instance = Environment(vehicle=Multirotor(quad_params),           # vehicle object, must be specified. 
                            controller=SE3Control(quad_params),        # controller object, must be specified.
                            trajectory=HoverTraj(x0=[0, 0, 0]),         # trajectory object, must be specified.
                            wind_profile=SinusoidWind(),               # OPTIONAL: wind profile object, if none is supplied it will choose no wind. 
                            sim_rate     = 100,                        # OPTIONAL: The update frequency of the simulator in Hz. Default is 100 Hz.
                            world        = world,                      # OPTIONAL: the world, same name as the file in rotorpy/worlds/, default (None) is empty world
                            safety_margin= 0.25                        # OPTIONAL: defines the radius (in meters) of the sphere used for collision checking
                        )

    # Setting an initial state. 
    print("\tSetting initial state...")
    x0 = {'x': np.array([0,0,0]),
        'v': np.zeros(3,),
        'q': np.array([0, 0, 0, 1]), # [i,j,k,w]
        'w': np.zeros(3,),
        'wind': np.array([0,0,0]),  # Since wind is handled elsewhere, this value is overwritten
        'rotor_speeds': np.array([1788.53, 1788.53, 1788.53, 1788.53])}
    sim_instance.vehicle.initial_state = x0

    # Running the simulation
    print("\tRunning the simulation...")
    results = sim_instance.run(t_final      = 1,        # The maximum duration of the environment in seconds
                            use_mocap    = False,       # Boolean: determines if the controller should use the motion capture estimates. 
                            terminate    = False,       # Boolean: if this is true, the simulator will terminate when it reaches the last waypoint.
                            plot            = True,     # Boolean: plots the vehicle states and commands   
                            plot_mocap      = True,     # Boolean: plots the motion capture pose and twist measurements
                            plot_estimator  = True,     # Boolean: plots the estimator filter states and covariance diagonal elements
                            plot_imu        = True,     # Boolean: plots the IMU measurements
                            animate_bool    = True,     # Boolean: determines if the animation of vehicle state will play. 
                            animate_wind    = True,     # Boolean: determines if the animation will include a scaled wind vector to indicate the local wind acting on the UAV. 
                            verbose         = True,     # Boolean: will print statistics regarding the simulation. 
                            fname   = None              # Filename is specified if you want to save the animation. The save location is rotorpy/data_out/. 
                        )

    # Save the simulation data to a CSV file
    print("\tSaving simulation data to CSV...")
    sim_instance.save_to_csv("test_simulation.csv")

    # Instead of producing a CSV, you can manually unpack the dictionary into a Pandas DataFrame using the following:
    print("\tUnpacking simulation data...") 
    dataframe = unpack_sim_data(results)

    plt.close('all')


'''
Test the trajectory module. 
'''

import inspect
import numpy as np
import glob
import os
import importlib

# The expected signature for the `update` method
EXPECTED_UPDATE_SIGNATURE = {
    'args': ['self', 't'],  # Expected arguments, 'self' is implicit
    'return_type': 'dict'  # Expected return type (dict)
}

# Define the expected structure of the returned dictionary
EXPECTED_RETURN_DICT_KEYS = {
    'x': np.ndarray,        # Position vector (m)
    'x_dot': np.ndarray,    # Velocity vector (m/s)
    'x_ddot': np.ndarray,   # Acceleration vector (m/s^2)
    'x_dddot': np.ndarray,  # Jerk vector (m/s^3)
    'x_ddddot': np.ndarray, # Snap vector (m/s^4)
    'yaw': float,           # Yaw angle (rad)
    'yaw_dot': float,       # Yaw rate (rad/s)
    'yaw_ddot': float,      # Yaw acceleration (rad/s^2)
}

def check_update_signature(class_obj):
    # Ensure the class has an 'update' method
    assert hasattr(class_obj, 'update'), f"{class_obj.__name__} does not have an 'update' method."

    # Get the signature of the 'update' method
    update_method = getattr(class_obj, 'update')

    # Get the signature of the method using inspect
    signature = inspect.signature(update_method)

    # Get the parameter names and return type
    params = list(signature.parameters.keys())  # Get parameter names as a list

    # Check if the parameter names match
    assert params == EXPECTED_UPDATE_SIGNATURE['args'], f"Expected arguments {EXPECTED_UPDATE_SIGNATURE['args']} for {class_obj.__name__}, but got {params}"

# Test case to check all trajectories adhere to input requirement. 
def test_trajectories_update_method():

    # Import all trajectories from the trajectories module
    traj_files = glob.glob(os.path.join(os.path.dirname(__file__), '..', 'rotorpy', 'trajectories', '*.py'))
    traj_objects = []
    print("\nTesting input signatures for all trajectory classes...")
    for file in traj_files:
        traj_name = os.path.basename(file)[:-3]  # Remove '.py' extension
        if traj_name != '__init__': # Ignore the __init__.py file
            print(f"Importing trajectory classes from {traj_name}")
            traj_module = importlib.import_module(f"rotorpy.trajectories.{traj_name}")  # Import the module
            traj_classes = inspect.getmembers(traj_module, inspect.isclass)            # Get all classes in the module

            # Ignore any dependency classes in the trajectories module
            traj_objects.extend([traj_class for traj_class in traj_classes if traj_class[1].__module__ == f"rotorpy.trajectories.{traj_name}"])

    # Iterate over all imported classes from the trajectories module
    for traj_class in traj_objects:
        print(f"Checking {traj_class[0]} input signatures...")
        if hasattr(traj_class[1], 'update'):  # Check only classes with an 'update' method
            check_update_signature(traj_class[1])

# Test MinSnap trajectory class. 
def test_minsnap_traj():
    from rotorpy.trajectories.minsnap import MinSnap

    print("\nTesting MinSnap class input and output signatures")
    # Create an instance of the MinSnap class
    min_snap = MinSnap(points=np.array([[0, 0, 0], [1, 2, 4], [3, 5, 2]]), 
                       yaw_angles=np.array([0, np.pi/4, np.pi/2]), 
                       yaw_rate_max=2*np.pi, 
                       poly_degree=7, 
                       yaw_poly_degree=7, 
                       v_max=3, 
                       v_avg=1, 
                       v_start=[0, 0, 0], 
                       v_end=[0, 0, 0], 
                       verbose=True)
    
    return_value = min_snap.update(0.0)
    assert isinstance(return_value, dict), "update() method should return a dictionary"

    # Check if the return value has the expected keys and types
    for key, value in EXPECTED_RETURN_DICT_KEYS.items():
        assert key in return_value, f"Expected key '{key}' in return dictionary"
        assert isinstance(return_value[key], value), f"Expected type {value} for key '{key}', but got {type(return_value[key])}"


import numpy as np
import torch

from rotorpy.trajectories.minsnap import BatchedMinSnap
from rotorpy.vehicles.multirotor import Multirotor, BatchedMultirotorParams, BatchedMultirotor
from rotorpy.controllers.quadrotor_control import BatchedSE3Control, SE3Control
from rotorpy.vehicles.crazyflie_params import quad_params as cf_quad_params
from rotorpy.vehicles.hummingbird_params import quad_params as hb_quad_params
from rotorpy.utils.trajgen_utils import generate_random_minsnap_traj
from rotorpy.world import World

def test_batched_operators():
    np.random.seed(10)
    num_drones = 10
    device = torch.device("cpu")
    init_rotor_speed = 1788.53
    # Set to 0 if you want sim results to be more deterministic (default value is 100)
    cf_quad_params["motor_noise_std"] = 0
    hb_quad_params["motor_noise_std"] = 0

    # We'll simulate half crazyflies, half hummingbirds
    all_quad_params = [cf_quad_params]*(num_drones//2) + [hb_quad_params]*(num_drones//2)

    world = World({"bounds": {"extents": [-10, 10, -10, 10, -10, 10]}, "blocks": []})
    batch_state = {'x': torch.randn(num_drones,3, device=device).double(),
                  'v': torch.randn(num_drones, 3, device=device).double(),
                  'q': torch.tensor([0, 0, 0, 1], device=device).repeat(num_drones, 1).double(),
                  'w': torch.randn(num_drones, 3, device=device).double(),
                  'wind': torch.zeros(num_drones, 3, device=device).double(),
                  'rotor_speeds': torch.tensor([init_rotor_speed, init_rotor_speed, init_rotor_speed, init_rotor_speed], device=device).repeat(num_drones, 1).double()}
    trajectories = [generate_random_minsnap_traj(world, 3, 1.0, 1.0, 2.0, np.random.randn(3)) for _ in range(num_drones)]
    batch_minsnap = BatchedMinSnap(trajectories, device)
    ts = np.array([np.random.uniform(0, 1) for i in range(num_drones)])
    batch_flat_output= batch_minsnap.update(ts)

    # Make sure that BatchedMinSnap does the same thing as MinSnap
    for j, traj in enumerate(trajectories):
        seq_flat_output = traj.update(ts[j])
        for key in batch_flat_output.keys():
            assert np.all(np.abs(batch_flat_output[key][j].cpu().numpy() - seq_flat_output[key]) < 1e-3)

    batch_params = BatchedMultirotorParams(all_quad_params, num_drones, device)
    batched_ctrlr = BatchedSE3Control(batch_params, num_drones, device)
    batch_control_inputs = batched_ctrlr.update(0, batch_state, batch_flat_output)

    # Make sure that BatchedSE3Control does the same thing as SE3Control
    # Make sure that BatchedMultirotor does the same thing as MultiRotor
    control_abstractions = ["cmd_motor_speeds", "cmd_motor_thrusts", "cmd_ctbm", "cmd_ctbr",
                            "cmd_ctatt", "cmd_vel", "cmd_acc"]
    for j in range(num_drones):
        single_ctrlr = SE3Control(all_quad_params[j])
        seq_state = {key: batch_state[key][j].cpu().numpy() for key in batch_state.keys()}
        flat_output = {key: batch_flat_output[key][j].cpu().numpy() for key in batch_flat_output.keys()}
        seq_control_input = single_ctrlr.update(0, seq_state, flat_output)
        for key in batch_control_inputs.keys():
            assert np.all(np.abs(batch_control_inputs[key][j].cpu().numpy() - seq_control_input[key]) < 1e-3)

    for abstraction in control_abstractions:
        print(f"Testing control abstraction = {abstraction}")
        batched_multirotor = BatchedMultirotor(batch_params, num_drones, batch_state, device, control_abstraction=abstraction)
        batch_next_state = batched_multirotor.step(batch_state, batch_control_inputs, 0.01)
        for j in range(num_drones):
            single_ctrlr = SE3Control(all_quad_params[j])
            seq_state = {key: batch_state[key][j].cpu().numpy() for key in batch_state.keys()}
            flat_output = {key: batch_flat_output[key][j].cpu().numpy() for key in batch_flat_output.keys()}
            seq_control_input = single_ctrlr.update(0, seq_state, flat_output)
            single_multirotor = Multirotor(all_quad_params[j], control_abstraction=abstraction)
            seq_next_state = single_multirotor.step(seq_state, seq_control_input, 0.01)
            for key in batch_next_state.keys():
                # since rotor speeds are large, we need a higher tolerance here.
                if key == "rotor_speeds":
                    assert np.all(np.abs(batch_next_state[key][j].cpu().numpy() - seq_next_state[key]) < 1)
                else:
                    assert np.all(np.abs(batch_next_state[key][j].cpu().numpy() - seq_next_state[key]) < 3e-2)


if __name__ == "__main__":
    test_batched_operators()



from foundation_policy import Raptor
from scipy.spatial.transform import Rotation
import numpy as np

class RaptorFoundationPolicy(object):
    """
    implementing the RAPTOP policy
    """
    def __init__(self, quad_params):
        """
        Parameters:
            quad_params, dict with keys specified in rotorpy/vehicles
        """

        self.policy = Raptor()
        self.policy.reset()

        # initialize the action
        self.action_past = np.zeros(4)

        # load rpm info
        self.rotor_speed_min = quad_params['rotor_speed_min']
        self.rotor_speed_max = quad_params['rotor_speed_max']

    def update(self, t, state, flat_output):
        """
        This function receives the current time, true state, and desired flat
        outputs. It returns the command inputs.

        Inputs:
            t, present time in seconds
            state, a dict describing the present state with keys
                x, position, m
                v, linear velocity, m/s
                q, quaternion [i,j,k,w]
                w, angular velocity, rad/s
            flat_output, a dict describing the present desired flat outputs with keys
                x,        position, m
                x_dot,    velocity, m/s
                x_ddot,   acceleration, m/s**2
                x_dddot,  jerk, m/s**3
                x_ddddot, snap, m/s**4
                yaw,      yaw angle, rad
                yaw_dot,  yaw rate, rad/s

        Outputs:
            control_input, a dict describing the present computed control inputs with keys
                cmd_motor_speeds, rad/s
                cmd_motor_thrusts, N
                cmd_thrust, N 
                cmd_moment, N*m
                cmd_q, quaternion [i,j,k,w]
                cmd_w, angular rates in the body frame, rad/s
                cmd_v, velocity in the world frame, m/s
        """

        # fetch the state estimates (the policy works under ENU frame)
        pos = state['x']
        vel = state['v']
        R = Rotation.from_quat(state['q']).as_matrix()
        omega = state['w']

        # fetch target pos and vel
        target_pos = flat_output['x']
        target_vel = flat_output['x_dot']
        
        # for trajectory, use pos-target_pos and vel-target_vel
        observation = np.hstack((pos - target_pos, R.flatten(), vel-target_vel, omega, self.action_past))

        # clip policy outputs to be the defined range [-1, 1]
        action = np.clip(self.policy.evaluate_step(observation)[0], -1, 1)
        # save current control action
        self.action_past = action

        # put dummy cmd_moment and cmd_q just to pass the sanity check
        # control-wise, the cmd_motor_speeds has priority
        cmd_moment_dummy = np.zeros((3,))
        cmd_q_dummy = np.array([0,0,0,1])
        cmd_thrust_dummy = 0
        
        # remap the motor numbers
        action_remapped = np.array([action[3], action[0], action[1], action[2]])
        
        # denormalize based on https://github.com/rl-tools/raptor?tab=readme-ov-file#usage
        action_rpm = (self.rotor_speed_max - self.rotor_speed_min) * (action_remapped + 1)/2 + self.rotor_speed_min
        
        # load control
        control_input = {'cmd_motor_speeds':action_rpm,
                         'cmd_thrust':cmd_thrust_dummy,
                         'cmd_moment':cmd_moment_dummy,
                         'cmd_q':cmd_q_dummy}
        
        return control_input


import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import StateSpace, lsim
from matplotlib.widgets import Slider, TextBox
from rotorpy.vehicles.crazyfliebrushless_params import quad_params, d

""" 
The point of this script is to provide an interactive tool to gain insight into approximately how the lower level attitude gains
(kp_att and kd_att) affect the controller response (control effort, nat freq, damping ratio, settling time) for the closed loop attitude tracking 
using a simplified model of the attitude dynamics.
"""
def interactive_pd_homogeneous_fixed():
    def compute_response(kp, kd, x0, inertia):
        """ 
        Given set values for the gains, the initial angle error (rad), and inertia, simulate the 
        attitude dynamics using a linear second order system. 
        """ 
        if kp <= 0 or kd < 0 or inertia <= 0:
            return None, None, None, None, None

        wn = np.sqrt(kp)
        zeta = kd / (2 * wn)
        wd = wn * np.sqrt(1 - zeta**2) if zeta < 1 else 0

        # State-space: x1 = position, x2 = velocity
        A = np.array([[0, 1],
                      [-kp, -kd]])
        B = np.array([[0], [0]])  # No external input
        C = np.array([[1, 0]])
        D = np.array([[0]])

        sys = StateSpace(A, B, C, D)
        t = np.linspace(0, 0.5, 1000)
        u = np.zeros_like(t)  # no external input
        x0_vec = [x0, 0]  # position error, zero initial velocity

        _, y, x = lsim(sys, U=u, T=t, X0=x0_vec)

        return wn, zeta, wd, t, y.squeeze(), x

    def compute_settling_time(t, y, threshold=0.02):
        """
        Compute the time it takes for the response to settle based on the 2% rule.  
        """
        if len(y) == 0:
            return t[-1]
        steady_state = 0.0
        within_bounds = np.abs(y - steady_state) <= threshold * np.abs(y[0])
        for i in reversed(range(len(y))):
            if not within_bounds[i]:
                return t[i + 1] if i + 1 < len(t) else t[-1]
        return t[-1]

    def compute_control_effort(kp, kd, x, inertia):
        """ 
        Back out the expected control effort from this controller based on the angle and body rate errors. 
        """ 
        u = -inertia * (kp * x[:, 0] + kd * x[:, 1])
        return u

    # Initial values (start with nominal CFBL gains)
    kp0 = 2624.0
    kd0 = 360.0
    x0 = 0.785398  # 45 deg error in attitude
    inertia0 = quad_params['Ixx']  # Inertia in the roll axis
    max_moment = 2*quad_params['k_eta']*(quad_params['rotor_speed_max']**2)*d*0.7071  # Maximum roll moment from quad params. Assuming 2 motors are at max speed and the other 2 are off. 

    # Get the system response and characteristics based on the initial gains/error. 
    wn, zeta, wd, t, y, x = compute_response(kp0, kd0, x0, inertia0)
    settling_time = compute_settling_time(t, y)
    u = compute_control_effort(kp0, kd0, x, inertia0)

    # Plot setup
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(14, 10))
    plt.subplots_adjust(bottom=0.3)

    response_line, = ax1.plot(t, y, lw=2, label="Homogeneous Response")
    st_line = ax1.axvline(settling_time, color='r', linestyle='--', label="Settling Time")
    info_text = ax1.text(0.02, 0.95, '', transform=ax1.transAxes, verticalalignment='top',
                         fontsize=10, bbox=dict(boxstyle="round", facecolor="white", alpha=0.7))
    ax1.set_title("PD Homogeneous Response (Initial Error)")
    ax1.set_xlabel("Time [s]")
    ax1.set_ylabel("Position")
    ax1.grid(True)
    ax1.legend(loc="lower right")

    effort_line, = ax2.plot(t, u, color='orange', lw=2, label="Control Effort")
    ax2.axhline(max_moment, color='gray', linestyle='--', label="Max Moment")
    ax2.axhline(-max_moment, color='gray', linestyle='--')
    ax2.set_title("Control Effort")
    ax2.set_xlabel("Time [s]")
    ax2.set_ylabel("Torque")
    ax2.grid(True)
    ax2.legend(loc="upper right")

    def update_info_text(wn, zeta, wd):
        info_text.set_text(
            f"Natural Freq (wn): {wn:.2f} rad/s\n"
            f"Damping Ratio (ζ): {zeta:.2f}\n"
            f"Damped Freq (wd): {wd:.2f} rad/s"
        )

    update_info_text(wn, zeta, wd)

    # Sliders and TextBox
    ax_kp = plt.axes([0.15, 0.22, 0.65, 0.03])
    ax_kd = plt.axes([0.15, 0.17, 0.65, 0.03])
    ax_x0 = plt.axes([0.15, 0.12, 0.65, 0.03])
    s_kp = Slider(ax_kp, 'kp', 100.0, 5000.0, valinit=kp0, valstep=1.0)
    s_kd = Slider(ax_kd, 'kd', 5.0, 1000.0, valinit=kd0, valstep=1.0)
    s_x0 = Slider(ax_x0, 'Initial Error', 0.01, 5.0, valinit=x0)

    ax_inertia = plt.axes([0.15, 0.05, 0.2, 0.035])
    box_inertia = TextBox(ax_inertia, 'Inertia', initial=str(inertia0))

    def update(val=None):
        """ 
        Update the plots whenever the sliders or text box are changed. 
        """ 
        try:
            inertia = float(box_inertia.text)
        except ValueError:
            inertia = inertia0

        kp = s_kp.val
        kd = s_kd.val
        x0 = s_x0.val

        wn, zeta, wd, t, y, x = compute_response(kp, kd, x0, inertia)
        if t is None:
            return

        u = compute_control_effort(kp, kd, x, inertia)
        settling_time = compute_settling_time(t, y)

        response_line.set_data(t, y)
        effort_line.set_data(t, u)
        st_line.set_xdata(settling_time)

        ax1.set_xlim([-0.05, t[-1]])
        ax1.set_ylim([min(y) - 0.1, max(y) + 0.1])
        ax2.set_xlim([-0.05, t[-1]])
        ax2.set_ylim([min(u) * 1.1, max(u) * 1.1])

        update_info_text(wn, zeta, wd)
        fig.canvas.draw_idle()

    s_kp.on_changed(update)
    s_kd.on_changed(update)
    s_x0.on_changed(update)
    box_inertia.on_submit(lambda val: update())

    plt.show()

if __name__ == "__main__":
    interactive_pd_homogeneous_fixed()